# coding=utf-8
# @license
# Copyright 2019-2020 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Wrappers for representing a Neuroglancer coordinate space."""

import collections
import numpy as np

__all__ = []


def export(obj):
    __all__.append(obj.__name__)
    return obj


si_prefixes = {
    'Y': 24,
    'Z': 21,
    'E': 18,
    'P': 15,
    'T': 12,
    'G': 9,
    'M': 6,
    'k': 3,
    'h': 2,
    '': 0,
    'c': -2,
    'm': -3,
    'u': -6,
    'µ': -6,
    'n': -9,
    'p': -12,
    'f': -15,
    'a': -18,
    'z': -21,
    'y': -24,
}

si_units = ['m', 's', 'rad/s', 'Hz']

si_units_with_prefixes = {
    '%s%s' % (prefix, unit): (unit, exponent)
    for (prefix, exponent) in si_prefixes.items() for unit in si_units
}

si_units_with_prefixes[''] = ('', 0)


def parse_unit(scale, unit):
    unit, exponent = si_units_with_prefixes[unit]
    if exponent >= 0:
        return (scale * 10**exponent, unit)
    else:
        return (scale / 10**(-exponent), unit)


@export
class CoordinateArray:
    __slots__ = ('_data')

    def __init__(self, json_data=None, labels=None, coordinates=None, mappings=None):
        if mappings is None:
            mappings = dict()
        else:
            mappings = dict(mappings)
        if labels is not None:
            if coordinates is None:
                coordinates = range(len(labels))
            for coordinate, label in zip(coordinates, labels):
                mappings[coordinate] = label
        if json_data is not None:
            if not isinstance(json_data,
                              dict) or 'coordinates' not in json_data or 'labels' not in json_data:
                raise ValueError('Expected object with "coordinates" and "labels" properties')
            coordinates = json_data['coordinates']
            labels = json_data['labels']
            for coordinate, label in zip(coordinates, labels):
                mappings[coordinate] = label
        self._data = mappings

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return repr(self._data)

    def __str__(self):
        return str(self._data)

    def __eq__(self, other):
        if not isinstance(other, CoordinateArray): return False
        return self._data == other._data

    def __getitem__(self, k):
        if isinstance(k, str):
            for other_k, other_v in self._data.items():
                if other_k == k: return other_v
            raise KeyError('label not found: %r' % (k, ))
        return self._data[k]

    def to_json(self):
        return dict(coordinates=list(self._data.keys()), labels=list(self._data.values()))


@export
class DimensionScale(collections.namedtuple('DimensionScale',
                                            ['scale', 'unit', 'coordinate_array'])):
    __slots__ = ()

    def __new__(cls, scale=1, unit='', coordinate_array=None):
        return super(DimensionScale, cls).__new__(cls, scale, unit, coordinate_array)

    @staticmethod
    def from_json(json):
        if isinstance(json, DimensionScale):
            return json
        if isinstance(json, list):
            if len(json) != 2:
                raise ValueError('Expected [scale, unit], but received: %r' % (json, ))
            scale = json[0]
            unit = json[1]
            coordinate_array = None
        else:
            scale = None
            unit = None
            coordinate_array = CoordinateArray(json_data=json)
        return DimensionScale(scale=scale, unit=unit, coordinate_array=coordinate_array)


@export
class CoordinateSpace(object):
    __slots__ = ('names', 'scales', 'units', 'coordinate_arrays')

    def __init__(self, json=None, names=None, scales=None, units=None, coordinate_arrays=None):
        if json is None:
            if names is not None:
                self.names = tuple(names)
                scales = np.array(scales, dtype=np.float64)
                if isinstance(units, str):
                    units = tuple(units for _ in names)
                scales_and_units = tuple(
                    parse_unit(scale, unit) for scale, unit in zip(scales, units))
                scales = np.array([s[0] for s in scales_and_units], dtype=np.float64)
                units = tuple(s[1] for s in scales_and_units)
                if coordinate_arrays is None:
                    coordinate_arrays = tuple(None for x in units)
                else:
                    coordinate_arrays = tuple(coordinate_arrays)
                self.units = units
                self.scales = scales
                self.coordinate_arrays = coordinate_arrays
            else:
                self.names = ()
                self.scales = np.zeros(0, dtype=np.float64)
                self.units = ()
                self.coordinate_arrays = ()
        else:
            if not isinstance(json, dict): raise TypeError
            self.names = tuple(json.keys())
            values = tuple(DimensionScale.from_json(v) for v in json.values())
            self.scales = np.array([v.scale for v in values], dtype=np.float64)
            self.units = tuple(v.unit for v in values)
            self.coordinate_arrays = tuple(v.coordinate_array for v in values)
        self.scales.setflags(write=False)

    @property
    def rank(self):
        return len(self.names)

    def __getitem__(self, i):
        if isinstance(i, str):
            idx = self.names.index(i)
            return DimensionScale(scale=self.scales[idx],
                                  unit=self.units[idx],
                                  coordinate_array=self.coordinate_arrays[idx])
        if isinstance(i, slice):
            idxs = range(self.rank)[i]
            return [
                DimensionScale(scale=self.scales[j],
                               unit=self.units[j],
                               coordinate_array=self.coordinate_arrays[j]) for j in idxs
            ]
        return DimensionScale(scale=self.scales[i],
                              unit=self.units[i],
                              coordinate_array=self.coordinate_arrays[i])

    def __repr__(self):
        return 'CoordinateSpace(%r)' % (self.to_json(), )

    def to_json(self):
        d = collections.OrderedDict()
        for name, scale, unit, coordinate_array in zip(self.names, self.scales, self.units,
                                                       self.coordinate_arrays):
            if coordinate_array is None:
                d[name] = [scale, unit]
            else:
                d[name] = coordinate_array.to_json()
        return d



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import concurrent.futures
import logging
import threading

from . import credentials_provider
from .futures import run_on_new_thread


class GoogleOAuth2FlowCredentialsProvider(credentials_provider.CredentialsProvider):
    def __init__(self, scopes, client_id, client_secret):
        super(GoogleOAuth2FlowCredentialsProvider, self).__init__()
        self.scopes = scopes
        self.client_id = client_id
        self.client_secret = client_secret

        # Make sure logging is initialized.  Does nothing if logging has already
        # been initialized.
        logging.basicConfig()

    def get_new(self):
        def func():
            import apitools.base.py.credentials_lib
            result = apitools.base.py.credentials_lib.GetCredentials(
                package_name='',
                scopes=self.scopes,
                client_id=self.client_id,
                client_secret=self.client_secret,
                user_agent=u'python-neuroglancer',
            )
            return dict(tokenType=u'Bearer', accessToken=result.get_access_token().access_token)

        return run_on_new_thread(func)


class GoogleApplicationDefaultCredentialsProvider(credentials_provider.CredentialsProvider):
    def __init__(self):
        super(GoogleApplicationDefaultCredentialsProvider, self).__init__()

        # Make sure logging is initialized.  Does nothing if logging has already
        # been initialized.
        logging.basicConfig()

        self._lock = threading.Lock()
        self._credentials = None

    def get_new(self):
        def func():
            with self._lock:
                if self._credentials is None:
                    import google.auth
                    credentials, project = google.auth.default()
                    del project
                    self._credentials = credentials
                if not self._credentials.valid:
                    import google.auth.transport.requests
                    import requests
                    request = google.auth.transport.requests.Request()
                    self._credentials.refresh(request)
                return dict(tokenType=u'Bearer', accessToken=self._credentials.token)

        return run_on_new_thread(func)


_global_google_application_default_credentials_provider = None
_global_google_application_default_credentials_provider_lock = threading.Lock()


def get_google_application_default_credentials_provider():
    global _global_google_application_default_credentials_provider
    with _global_google_application_default_credentials_provider_lock:
        if _global_google_application_default_credentials_provider is None:
            _global_google_application_default_credentials_provider = GoogleApplicationDefaultCredentialsProvider(
            )
        return _global_google_application_default_credentials_provider



#!/usr/bin/env python3

import argparse
import os
import shutil
import subprocess
import sys

script_dir = os.path.dirname(sys.argv[0])
src_dir = os.path.join(script_dir, 'ext/src')
dest_openmesh_dir = os.path.join(script_dir, 'ext/third_party/openmesh/OpenMesh')

ap = argparse.ArgumentParser()
ap.add_argument('openmesh_directory', help='Path to OpenMesh root directory')
args = ap.parse_args()

openmesh_dir = os.path.abspath(args.openmesh_directory)

deps = subprocess.check_output(
    ['gcc', '-pthread', '-I', os.path.join(openmesh_dir, 'src'), '-c', 'openmesh_dependencies.cc',
     'on_demand_object_mesh_generator.cc', '-MM', '-MF', '/dev/stdout',
     '-fopenmp', '-std=c++11'], cwd=src_dir).split()
deps = [x[len(openmesh_dir)+1:] for x in deps if x.startswith(openmesh_dir + '/')] + ['LICENSE', 'VERSION']

if os.path.exists(dest_openmesh_dir):
  shutil.rmtree(dest_openmesh_dir)

for dep in deps:
  dest_path = os.path.join(dest_openmesh_dir, dep)
  dest_dir = os.path.dirname(dest_path)
  if not os.path.exists(dest_dir):
    os.makedirs(dest_dir)
  shutil.copyfile(os.path.join(openmesh_dir, dep), dest_path)



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various extensions to the concurrent.futures module."""

from __future__ import absolute_import

import concurrent.futures
import threading


def future_then_immediate(future, func, new_future = None):
    """Returns a future that maps the result of `future` by `func`.

    If `future` succeeds, sets the result of the returned future to `func(future.result())`.  If
    `future` fails or `func` raises an exception, the exception is stored in the returned future.

    If `future` has not yet finished, `func` is invoked by the same thread that finishes it.
    Otherwise, it is invoked immediately in the same thread that calls `future_then_immediate`.
    """
    if new_future is None:
        new_future = concurrent.futures.Future()

    def on_done(f):
        try:
            new_future.set_result(func(f.result()))
        except Exception as e:
            new_future.set_exception(e)

    future.add_done_callback(on_done)
    return new_future


def run_on_new_thread(func, daemon=True):
    """Calls `func()` from a new thread.

    :param daemon: Specifies whether the thread should be in daemon mode.

    :returns: A concurrent.futures.Future object representing the result.
    """
    f = concurrent.futures.Future()
    def wrapper():
        if not f.set_running_or_notify_cancel():
            return
        try:
            f.set_result(func())
        except Exception as e:
            f.set_exception(e)
    t = threading.Thread(target=wrapper)
    t.daemon = daemon
    t.start()
    return f



# @license
# Copyright 2021 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module implements function for authentication of layers based on DVID.
    Here tokens are fetched from local locations like env vars etc."""

from __future__ import absolute_import

import logging
import os

from . import credentials_provider
from .futures import run_on_new_thread


class TokenbasedDefaultCredentialsProvider(credentials_provider.CredentialsProvider):
    def __init__(self, parameters):
        super(TokenbasedDefaultCredentialsProvider, self).__init__()

        # Make sure logging is initialized.
        # Does nothing if logging has already been initialized.
        logging.basicConfig()
        self.parameters = parameters
        self._credentials = {}

    def get_new(self):
        def func():
            try:
                credentials = os.environ['DVID_CREDENTIALS']
                credentials = dict(item.split("=") for item in credentials.split(","))
                token = credentials[self.parameters['dvidServer']]
            except KeyError:
                raise RuntimeError(
                    """DVID_CREDENTIALS is not defined in your environment or does
                    not contain the token for the server: """ +
                    self.parameters['dvidServer'])
            self._credentials['token'] = token
            return dict(tokenType=u'Bearer', accessToken=self._credentials['token'])

        return run_on_new_thread(func)


_global_tokenbased_application_default_credentials_provider = None


def get_tokenbased_application_default_credentials_provider(parameters):
    global _global_tokenbased_application_default_credentials_provider
    if _global_tokenbased_application_default_credentials_provider is None:
        _global_tokenbased_application_default_credentials_provider =\
            TokenbasedDefaultCredentialsProvider(parameters)
    return _global_tokenbased_application_default_credentials_provider



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import concurrent.futures
import json
import multiprocessing
import re
import socket
import sys
import threading
import weakref

import numpy as np
import tornado.httpserver
import tornado.netutil
import tornado.platform.asyncio
import tornado.web

from . import local_volume, static
from . import skeleton
from .json_utils import json_encoder_default, encode_json
from .random_token import make_random_token

INFO_PATH_REGEX = r'^/neuroglancer/info/(?P<token>[^/]+)$'
SKELETON_INFO_PATH_REGEX = r'^/neuroglancer/skeletoninfo/(?P<token>[^/]+)$'

DATA_PATH_REGEX = r'^/neuroglancer/(?P<data_format>[^/]+)/(?P<token>[^/]+)/(?P<scale_key>[^/]+)/(?P<start>[0-9]+(?:,[0-9]+)*)/(?P<end>[0-9]+(?:,[0-9]+)*)$'

SKELETON_PATH_REGEX = r'^/neuroglancer/skeleton/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

MESH_PATH_REGEX = r'^/neuroglancer/mesh/(?P<key>[^/]+)/(?P<object_id>[0-9]+)$'

STATIC_PATH_REGEX = r'^/v/(?P<viewer_token>[^/]+)/(?P<path>(?:[a-zA-Z0-9_\-][a-zA-Z0-9_\-.]*)?)$'

ACTION_PATH_REGEX = r'^/action/(?P<viewer_token>[^/]+)$'

EVENTS_PATH_REGEX = r'^/events/(?P<viewer_token>[^/]+)$'

SET_STATE_PATH_REGEX = r'^/state/(?P<viewer_token>[^/]+)$'

CREDENTIALS_PATH_REGEX = r'^/credentials/(?P<viewer_token>[^/]+)$'

global_static_content_source = None

global_server_args = dict(bind_address='127.0.0.1', bind_port=0)

debug = False

_IS_GOOGLE_COLAB = 'google.colab' in sys.modules


def _get_server_url(bind_address: str, port: int) -> str:
    if _IS_GOOGLE_COLAB:
        return _get_colab_server_url(port)
    return _get_regular_server_url(bind_address, port)


def _get_regular_server_url(bind_address: str, port: int) -> str:
    if bind_address == '0.0.0.0' or bind_address == '::':
        hostname = socket.getfqdn()
    else:
        hostname = bind_address
    return f'http://{hostname}:{port}'


def _get_colab_server_url(port: int) -> str:
    import google.colab.output
    return google.colab.output.eval_js(f'google.colab.kernel.proxyPort({port})')


class Server(object):
    def __init__(self, ioloop, bind_address='127.0.0.1', bind_port=0):
        self.viewers = weakref.WeakValueDictionary()
        self.token = make_random_token()
        self.executor = concurrent.futures.ThreadPoolExecutor(
            max_workers=multiprocessing.cpu_count())

        self.ioloop = ioloop

        def log_function(handler):
            if debug:
                print("%d %s %.2fs" %
                      (handler.get_status(), handler.request.uri, handler.request.request_time()))

        app = self.app = tornado.web.Application(
            [
                (STATIC_PATH_REGEX, StaticPathHandler, dict(server=self)),
                (INFO_PATH_REGEX, VolumeInfoHandler, dict(server=self)),
                (SKELETON_INFO_PATH_REGEX, SkeletonInfoHandler, dict(server=self)),
                (DATA_PATH_REGEX, SubvolumeHandler, dict(server=self)),
                (SKELETON_PATH_REGEX, SkeletonHandler, dict(server=self)),
                (MESH_PATH_REGEX, MeshHandler, dict(server=self)),
                (ACTION_PATH_REGEX, ActionHandler, dict(server=self)),
                (EVENTS_PATH_REGEX, EventStreamHandler, dict(server=self)),
                (SET_STATE_PATH_REGEX, SetStateHandler, dict(server=self)),
                (CREDENTIALS_PATH_REGEX, CredentialsHandler, dict(server=self)),
            ],
            log_function=log_function,
            # Set a large maximum message size to accommodate large screenshot
            # messages.
            websocket_max_message_size=100 * 1024 * 1024)
        http_server = tornado.httpserver.HTTPServer(
            app,
            # Allow very large requests to accommodate large screenshots.
            max_buffer_size=1024**3,
        )
        sockets = tornado.netutil.bind_sockets(port=bind_port, address=bind_address)
        http_server.add_sockets(sockets)
        actual_port = sockets[0].getsockname()[1]

        global global_static_content_source
        if global_static_content_source is None:
            global_static_content_source = static.get_default_static_content_source()
        self.port = actual_port
        self.server_url = _get_server_url(bind_address, actual_port)
        self.regular_server_url = _get_regular_server_url(bind_address, actual_port)
        self._credentials_manager = None

    def get_volume(self, key):
        dot_index = key.find('.')
        if dot_index == -1:
            return None
        viewer_token = key[:dot_index]
        volume_token = key[dot_index + 1:]
        viewer = self.viewers.get(viewer_token)
        if viewer is None:
            return None
        return viewer.volume_manager.volumes.get(volume_token)


class BaseRequestHandler(tornado.web.RequestHandler):
    def initialize(self, server):
        self.server = server


class StaticPathHandler(BaseRequestHandler):
    def get(self, viewer_token, path):
        if viewer_token != self.server.token and viewer_token not in self.server.viewers:
            self.send_error(404)
            return
        try:
            data, content_type = global_static_content_source.get(path)
        except ValueError as e:
            self.send_error(404, message=e.args[0])
            return
        self.set_header('Content-type', content_type)
        self.finish(data)


class ActionHandler(BaseRequestHandler):
    def post(self, viewer_token):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        action = json.loads(self.request.body)
        self.server.ioloop.add_callback(viewer.actions.invoke, action['action'], action['state'])
        self.finish('')


class EventStreamStateWatcher:
    def __init__(self, key: str, client_id: str, state, last_generation: str, wake_up):
        self.key = key
        self.state = state
        self.last_generation = last_generation
        self._wake_up = wake_up
        self._client_id = client_id
        state.add_changed_callback(wake_up)

    def unregister(self):
        self.state.remove_changed_callback(self._wake_up)

    def maybe_send_update(self, handler):
        raw_state, generation = self.state.raw_state_and_generation
        if generation == self.last_generation: return False
        if generation.startswith(self._client_id + '/'): return False
        self.last_generation = generation
        msg = {'k': self.key, 's': raw_state, 'g': generation}
        handler.write(f'data: {encode_json(msg)}\n\n')
        return True


class EventStreamHandler(BaseRequestHandler):
    async def get(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        self.set_header('content-type', 'text/event-stream')
        self.set_header('cache-control', 'no-cache')
        must_flush = True
        wake_event = asyncio.Event()
        self._wake_up = lambda: self.server.ioloop.add_callback(wake_event.set)
        client_id = self.get_query_argument('c')
        self._closed = False

        watchers = []

        def watch(key: str, state):
            watchers.append(
                EventStreamStateWatcher(key=key,
                                        state=state,
                                        client_id=client_id,
                                        wake_up=self._wake_up,
                                        last_generation=self.get_query_argument(f'g{key}')))

        watch('c', viewer.config_state)
        if hasattr(viewer, 'shared_state'):
            watch('s', viewer.shared_state)

        try:
            while True:
                wake_event.clear()
                if self._closed: break
                try:
                    sent = False
                    for watcher in watchers:
                        sent = watcher.maybe_send_update(self) or sent
                    if sent:
                        self.flush()
                        if _IS_GOOGLE_COLAB:
                            # The proxy used by colab buffers the entire
                            # response.  Therefore we must end the response for
                            # the client to receive it.
                            break
                    elif must_flush:
                        self.flush()
                except tornado.iostream.StreamClosedError:
                    break
                await wake_event.wait()
                must_flush = False
        finally:
            for watcher in watchers:
                watcher.unregister()

    def on_connection_close(self):
        if debug:
            print('connection closed')
        super().on_connection_close()
        self._wake_up()
        self._closed = True


class SetStateHandler(BaseRequestHandler):
    def post(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        msg = json.loads(self.request.body)
        generation = msg['g']
        state = msg['s']
        client_id = msg['c']
        viewer.set_state(state, f'{client_id}/{generation}')


class CredentialsHandler(BaseRequestHandler):
    async def post(self, viewer_token: str):
        viewer = self.server.viewers.get(viewer_token)
        if viewer is None:
            self.send_error(404)
            return
        if not viewer.allow_credentials:
            self.send_error(403)
            return
        if self.server._credentials_manager is None:
            from .default_credentials_manager import default_credentials_manager
            self.server._credentials_manager = default_credentials_manager
        msg = json.loads(self.request.body)
        invalid = msg.get('invalid')
        provider = self.server._credentials_manager.get(msg['key'], msg.get('parameters'))
        if provider is None:
            self.send_error(400)
            return
        try:
            credentials = await asyncio.wrap_future(provider.get(invalid))
            self.set_header('Content-type', 'application/json')
            self.finish(json.dumps(credentials))
        except:
            import traceback
            traceback.print_exc()
            self.send_error(401)


class VolumeInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())


class SkeletonInfoHandler(BaseRequestHandler):
    def get(self, token):
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return
        self.finish(json.dumps(vol.info(), default=json_encoder_default).encode())


class SubvolumeHandler(BaseRequestHandler):
    async def get(self, data_format, token, scale_key, start, end):
        start_pos = np.array(start.split(','), dtype=np.int64)
        end_pos = np.array(end.split(','), dtype=np.int64)
        vol = self.server.get_volume(token)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        try:
            data, content_type = await asyncio.wrap_future(
                self.server.executor.submit(vol.get_encoded_subvolume,
                                            data_format=data_format,
                                            start=start_pos,
                                            end=end_pos,
                                            scale_key=scale_key))
        except ValueError as e:
            self.send_error(400, message=e.args[0])
            return
        self.set_header('Content-type', content_type)
        self.finish(data)


class MeshHandler(BaseRequestHandler):
    async def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, local_volume.LocalVolume):
            self.send_error(404)
            return

        try:
            encoded_mesh = await asyncio.wrap_future(
                self.server.executor.submit(vol.get_object_mesh, object_id))
        except local_volume.MeshImplementationNotAvailable:
            self.send_error(501, message='Mesh implementation not available')
            return
        except local_volume.MeshesNotSupportedForVolume:
            self.send_error(405, message='Meshes not supported for volume')
            return
        except local_volume.InvalidObjectIdForMesh:
            self.send_error(404, message='Mesh not available for specified object id')
            return
        except ValueError as e:
            self.send_error(400, message=e.args[0])
            return

        self.set_header('Content-type', 'application/octet-stream')
        self.finish(encoded_mesh)


class SkeletonHandler(BaseRequestHandler):
    async def get(self, key, object_id):
        object_id = int(object_id)
        vol = self.server.get_volume(key)
        if vol is None or not isinstance(vol, skeleton.SkeletonSource):
            self.send_error(404)
            return

        def get_encoded_skeleton(skeletons, object_id):
            skeleton = skeletons.get_skeleton(object_id)
            if skeleton is None:
                return None
            return skeleton.encode(skeletons)

        try:
            encoded_skeleton = await asyncio.wrap_future(
                self.server.executor.submit(get_encoded_skeleton, vol, object_id))
        except:
            self.send_error(500, message=e.args[0])
            return
        if encoded_skeleton is None:
            self.send_error(404, message='Skeleton not available for specified object id')
            return
        self.set_header('Content-type', 'application/octet-stream')
        self.finish(encoded_skeleton)


global_server = None


def set_static_content_source(*args, **kwargs):
    global global_static_content_source
    global_static_content_source = static.get_static_content_source(*args, **kwargs)


def set_server_bind_address(bind_address='127.0.0.1', bind_port=0):
    global global_server_args
    global_server_args = dict(bind_address=bind_address, bind_port=bind_port)


def is_server_running():
    return global_server is not None


def stop():
    """Stop the server, invalidating any viewer URLs.

    This allows any previously-referenced data arrays to be garbage collected if there are no other
    references to them.
    """
    global global_server
    if global_server is not None:
        ioloop = global_server.ioloop

        def stop_ioloop():
            ioloop.stop()

        global_server.ioloop.add_callback(stop_ioloop)
        global_server = None


def get_server_url():
    return global_server.server_url


_global_server_lock = threading.Lock()


def start():
    global global_server
    with _global_server_lock:
        if global_server is not None: return

        # Workaround https://bugs.python.org/issue37373
        # https://www.tornadoweb.org/en/stable/index.html#installation
        if sys.platform == 'win32' and sys.version_info >= (3, 8):
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

        done = threading.Event()

        def start_server():
            global global_server
            ioloop = tornado.platform.asyncio.AsyncIOLoop()
            ioloop.make_current()
            asyncio.set_event_loop(ioloop.asyncio_loop)
            global_server = Server(ioloop=ioloop, **global_server_args)
            done.set()
            ioloop.start()
            ioloop.close()

        thread = threading.Thread(target=start_server)
        thread.daemon = True
        thread.start()
        done.wait()


def register_viewer(viewer):
    start()
    global_server.viewers[viewer.token] = viewer


def defer_callback(callback, *args, **kwargs):
    """Register `callback` to run in the server event loop thread."""
    start()
    global_server.ioloop.add_callback(lambda: callback(*args, **kwargs))



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import posixpath
import re


mime_type_map = {
    '.css': 'text/css',
    '.js': 'application/javascript',
    '.html': 'text/html',
    '.map': 'application/json'
}


def guess_mime_type_from_path(path):
    return mime_type_map.get(posixpath.splitext(path)[1], 'application/octet-stream')


class StaticContentSource(object):
    def get(self, name):
        if name == '':
            name = 'index.html'
        return self.get_content(name), guess_mime_type_from_path(name)

    def get_content(self, name):
        raise NotImplementedError


class PkgResourcesContentSource(StaticContentSource):
    def get_content(self, name):
        import pkg_resources
        if not re.match(r'^[a-z][a-z_\-\.]*\.(?:js|js\.map|css|html)$', name):
            raise ValueError('Invalid static resource name: %r' % name)
        if pkg_resources.resource_exists(__name__, name):
            return pkg_resources.resource_string(__name__, name)
        raise ValueError(
            'Static resources not built.  Run: "npm run build-python" or use an alternative static content source.'
        )


class HttpSource(StaticContentSource):
    def __init__(self, url):
        self.url = url

    def get_content(self, name):
        import requests
        full_url = posixpath.join(self.url, name)
        r = requests.get(full_url)
        if r.status_code >= 200 and r.status_code < 300:
            return r.content
        raise ValueError('Failed to retrieve %r: %s' % (full_url, r.reason))


class FileSource(StaticContentSource):
    def __init__(self, path, file_open=None):
        self.file_path = path
        self.file_open = file_open or open

    def get_content(self, name):
        full_path = os.path.join(self.file_path, name)
        try:
            with self.file_open(full_path, 'rb') as f:
                return f.read()
        except Exception as e:
            raise ValueError('Failed to read local path %r: %s' % (full_path, e))


def get_default_static_content_source():
    return PkgResourcesContentSource()


def get_static_content_source(source=None, url=None, path=None, file_open=None):
    if source is not None:
        return source
    elif url is not None:
        return HttpSource(url)
    elif path is not None:
        return FileSource(path=path, file_open=file_open)
    else:
        return get_default_static_content_source()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import copy

import six


class EquivalenceMap(object):
    """Union-find data structure"""

    supports_readonly = True

    def __init__(self, existing=None, _readonly=False):
        """Create a new empty union-find structure."""
        if isinstance(existing, EquivalenceMap):
            self._weights = existing._weights.copy()
            self._parents = existing._parents.copy()
            self._prev_next = existing._prev_next.copy()
            self._min_values = existing._min_values.copy()
        else:
            self._weights = {}
            self._parents = {}
            self._prev_next = {}
            self._min_values = {}
            self._readonly = False
            if existing is not None:
                if isinstance(existing, dict):
                    existing = six.viewitems(existing)
                for group in existing:
                    self.union(*group)
        self._readonly = _readonly

    def _get_representative(self, obj):
        """Finds and returns the root of the set containing `obj`."""

        if obj not in self._parents:
            self._parents[obj] = obj
            self._weights[obj] = 1
            self._prev_next[obj] = [obj, obj]
            self._min_values[obj] = obj
            return obj

        path = [obj]
        root = self._parents[obj]
        while root != path[-1]:
            path.append(root)
            root = self._parents[root]

        # compress the path and return
        for ancestor in path:
            self._parents[ancestor] = root
        return root

    def __getitem__(self, obj):
        """Returns the minimum element in the set containing `obj`."""
        if obj not in self._parents:
            return obj
        return self._min_values[self._get_representative(obj)]

    def __iter__(self):
        """Iterates over all elements known to this equivalence map."""
        return iter(self._parents)

    def items(self):
        return six.viewitems(self._parents)

    def keys(self):
        return six.viewkeys(self._parents)

    def clear(self):
        self._weights.clear()
        self._parents.clear()
        self._prev_next.clear()
        self._min_values.clear()

    def union(self, *args):
        """Unions the equivalence classes containing the elements in `*args`."""
        if self._readonly:
            raise AttributeError

        if len(args) == 0:
            return None
        if len(args) == 1:
            return self[args[0]]
        for a, b in zip(args[:-1], args[1:]):
            result = self._union_pair(a, b)
        return result

    def _union_pair(self, a, b):
        a = self._get_representative(a)
        b = self._get_representative(b)
        if a == b:
            return self._min_values[a]
        if self._weights[a] < self._weights[b]:
            a, b = (b, a)

        self._min_values[a] = min(self._min_values[a], self._min_values[b])

        a_links_new = self._prev_next[a]
        a_links_old = tuple(a_links_new)
        b_links_new = self._prev_next[b]
        b_links_old = tuple(b_links_new)

        # We want to splice b's list at the end of a's list

        # Splice beginning of b's list into end of a's list
        b_links_new[0] = a_links_old[0]
        self._prev_next[a_links_old[0]][1] = b

        # Splice end of b's list into end of a's list

        # last element of a's list is set to last element of b's list
        a_links_new[0] = b_links_old[0]
        # fix next pointer for last element of b's list
        self._prev_next[b_links_old[0]][1] = a
        self._weights[a] += self._weights[b]
        self._parents[b] = a
        return self._min_values[a]

    def members(self, x):
        """Yields the members of the equivalence class containing `x`."""
        if x not in self._parents:
            yield x
            return
        cur_x = x
        while True:
            yield cur_x
            cur_x = self._prev_next[cur_x][1]
            if cur_x == x:
                break

    def sets(self):
        """Returns the equivalence classes as a set of sets."""
        sets = {}
        for x in self._parents:
            sets.setdefault(self[x], set()).add(x)
        return frozenset(frozenset(v) for v in six.viewvalues(sets))

    def to_json(self):
        """Returns the equivalence classes a sorted list of sorted lists."""
        sets = self.sets()
        return sorted(sorted(x) for x in sets)

    def __copy__(self):
        """Does not preserve _readonly attribute."""
        return EquivalenceMap(self)

    def __deepcopy__(self, memo):
        """Does not preserve _readonly attribute."""
        result = EquivalenceMap()
        result._parents = copy.deepcopy(self._parents, memo)
        result._weights = copy.deepcopy(self._weights, memo)
        result._prev_next = copy.deepcopy(self._prev_next, memo)
        result._min_values = copy.deepcopy(self._min_values, memo)
        return result

    def copy(self):
        """Returns a copy of the equivalence map."""
        return EquivalenceMap(self)

    def delete_set(self, x):
        """Removes the equivalence class containing `x`."""
        if x not in self._parents:
            return
        members = list(self.members(x))
        for v in members:
            del self._parents[v]
            del self._weights[v]
            del self._prev_next[v]
            del self._min_values[v]

    def isolate_element(self, x):
        """Isolates `x` from its equivalence class."""
        members = list(self.members(x))
        self.delete_set(x)
        self.union(*(v for v in members if v != x))



#!/usr/bin/env python
import sys
if sys.version_info < (3, 5):
    print('Python >= 3.5 is required to build')
    sys.exit(1)

# Import setuptools before distutils because setuptools monkey patches
# distutils.
import setuptools

import atexit
import distutils.command.build
import os
import platform
import subprocess
import tempfile
import time
import setuptools.command.build_ext
import setuptools.command.develop
import setuptools.command.install
import setuptools.command.sdist
import setuptools.command.test

package_name = 'neuroglancer'
root_dir = os.path.dirname(__file__)
python_dir = os.path.join(root_dir, 'python')
src_dir = os.path.join(python_dir, 'ext', 'src')
openmesh_dir = os.path.join(python_dir, 'ext', 'third_party', 'openmesh', 'OpenMesh', 'src')

with open(os.path.join(python_dir, 'README.md'), mode='r', encoding='utf-8') as f:
    long_description = f.read()


def _maybe_bundle_client(cmd):
    """Build the client bundle if it does not already exist.

    If it has already been built but is stale, the user is responsible for
    rebuilding it.
    """

    bundle_client_cmd = cmd.distribution.get_command_obj('bundle_client')
    if bundle_client_cmd.skip_rebuild is None:
        bundle_client_cmd.skip_rebuild = True
    cmd.run_command('bundle_client')


def _setup_temp_egg_info(cmd):
    """Use a temporary directory for the `neuroglancer.egg-info` directory.

    When building an sdist (source distribution) or installing, locate the
    `neuroglancer.egg-info` directory inside a temporary directory so that it
    doesn't litter the source directory and doesn't pick up a stale SOURCES.txt
    from a previous build.
    """
    egg_info_cmd = cmd.distribution.get_command_obj('egg_info')
    if egg_info_cmd.egg_base is None:
        tempdir = tempfile.TemporaryDirectory(dir=os.curdir)
        egg_info_cmd.egg_base = tempdir.name
        atexit.register(tempdir.cleanup)


class SdistCommand(setuptools.command.sdist.sdist):
    def run(self):
        # Build the client bundle if it does not already exist.  If it has
        # already been built but is stale, the user is responsible for
        # rebuilding it.
        _maybe_bundle_client(self)
        _setup_temp_egg_info(self)
        super().run()

    def make_release_tree(self, base_dir, files):
        # Exclude .egg-info from source distribution.  These aren't actually
        # needed, and due to the use of the temporary directory in `run`, the
        # path isn't correct if it gets included.
        files = [x for x in files if '.egg-info' not in x]
        super().make_release_tree(base_dir, files)


class BuildCommand(distutils.command.build.build):
    def finalize_options(self):
        if self.build_base == 'build':
            # Use temporary directory instead, to avoid littering the source directory
            # with a `build` sub-directory.
            tempdir = tempfile.TemporaryDirectory()
            self.build_base = tempdir.name
            atexit.register(tempdir.cleanup)
        super().finalize_options()

    def run(self):
        _maybe_bundle_client(self)
        super().run()


class BuildExtCommand(setuptools.command.build_ext.build_ext):
    def finalize_options(self):
        super().finalize_options()
        # Prevent numpy from thinking it is still in its setup process
        if isinstance(__builtins__, dict):
            __builtins__['__NUMPY_SETUP__'] = False
        else:
            setattr(__builtins__, '__NUMPY_SETUP__', False)
        import numpy
        self.include_dirs.append(numpy.get_include())


class InstallCommand(setuptools.command.install.install):
    def run(self):
        _setup_temp_egg_info(self)
        super().run()


class DevelopCommand(setuptools.command.develop.develop):
    def run(self):
        _maybe_bundle_client(self)
        super().run()


class BundleClientCommand(distutils.command.build.build):

    user_options = [
        ('client-bundle-type=', None,
         'The nodejs bundle type. "min" (default) creates condensed static files for production, "dev" creates human-readable files.'
         ),
        ('skip-npm-reinstall', None,
         'Skip running `npm install` if the `node_modules` directory already exists.'),
        ('skip-rebuild', None,
         'Skip rebuilding if the `python/neuroglancer/static/index.html` file already exists.'),
    ]

    def initialize_options(self):

        self.client_bundle_type = 'min'
        self.skip_npm_reinstall = None
        self.skip_rebuild = None

    def finalize_options(self):

        if self.client_bundle_type not in ['min', 'dev']:
            raise RuntimeError('client-bundle-type has to be one of "min" or "dev"')

        if self.skip_npm_reinstall is None:
            self.skip_npm_reinstall = False

        if self.skip_rebuild is None:
            self.skip_rebuild = False

    def run(self):

        if self.skip_rebuild:
            html_path = os.path.join(python_dir, 'neuroglancer', 'static', 'index.html')
            if os.path.exists(html_path):
                print('Skipping rebuild of client bundle since %s already exists' % (html_path, ))
                return

        target = {"min": "build-python-min", "dev": "build-python-dev"}

        try:
            t = target[self.client_bundle_type]
            node_modules_path = os.path.join(root_dir, 'node_modules')
            if (self.skip_npm_reinstall and os.path.exists(node_modules_path)):
                print('Skipping `npm install` since %s already exists' % (node_modules_path, ))
            else:
                subprocess.call('npm i', shell=True, cwd=root_dir)
            res = subprocess.call('npm run %s' % t, shell=True, cwd=root_dir)
        except:
            raise RuntimeError(
                'Could not run \'npm run %s\'. Make sure node.js >= v12 is installed and in your path.'
                % t)

        if res != 0:
            raise RuntimeError('failed to bundle neuroglancer node.js project')


local_sources = [
    '_neuroglancer.cc',
    'openmesh_dependencies.cc',
    'on_demand_object_mesh_generator.cc',
    'voxel_mesh_generator.cc',
    'mesh_objects.cc',
]

USE_OMP = False
if USE_OMP:
    openmp_flags = ['-fopenmp']
else:
    openmp_flags = []

extra_compile_args = ['-std=c++11', '-fvisibility=hidden', '-O3'] + openmp_flags
if platform.system() == 'Darwin':
    extra_compile_args.insert(0, '-stdlib=libc++')

# Disable newer exception handling from Visual Studio 2019, since it requires a
# newer C++ runtime than shipped with Python.  The C++ extension doesn't use
# exceptions anyway.
#
# https://cibuildwheel.readthedocs.io/en/stable/faq/#importerror-dll-load-failed-the-specific-module-could-not-be-found-error-on-windows
if platform.system() == 'Windows':
    extra_compile_args.append('/d2FH4-')

# Copied from setuptools_scm, can be removed once a released version of
# setuptools_scm supports `version_scheme=no-guess-dev`.
#
# Note: It would be nice to include the commit hash in the version, but that
# can't be done in a PEP 440-compatible way.
def _no_guess_dev_version(version):
    if version.exact:
        return version.format_with("{tag}")
    else:
        return version.format_with("{tag}.post1.dev{distance}")


setuptools.setup(
    name=package_name,

    # Use setuptools_scm to determine version from git tags
    use_scm_version={
        "relative_to": __file__,
        "version_scheme": _no_guess_dev_version,
        "local_scheme": "no-local-version",
        "parentdir_prefix_version": package_name + "-",
    },
    description='Python data backend for neuroglancer, a WebGL-based viewer for volumetric data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jeremy Maitin-Shepard',
    author_email='jbms@google.com',
    url='https://github.com/google/neuroglancer',
    license='Apache License 2.0',
    packages=setuptools.find_packages('python'),
    package_dir={
        '': 'python',
    },
    package_data={
        'neuroglancer.static': ['*.html', '*.css', '*.js', '*.js.map'],
    },
    setup_requires=[
        "setuptools_scm>=4.1.2",
        "numpy>=1.11.0",
    ],
    install_requires=[
        "Pillow>=3.2.0",
        "numpy>=1.11.0",
        'requests',
        'tornado',
        'six',
        'google-apitools',
        'google-auth',
        'atomicwrites'
    ],
    extras_require={
        'test': [
            'pytest>=6.1.2',
            'pytest-rerunfailures>=9.1.1',
            'pytest-timeout>=1.4.2',
        ],
        'test-browser': [
            'selenium>=3.141.0',
            'webdriver-manager>=3.5.0',
        ],
    },
    ext_modules=[
        setuptools.Extension(
            'neuroglancer._neuroglancer',
            sources=[os.path.join(src_dir, name) for name in local_sources],
            language='c++',
            include_dirs=[openmesh_dir],
            define_macros=[
                ('_USE_MATH_DEFINES', None),  # Needed by OpenMesh when used with MSVC
            ],
            extra_compile_args=extra_compile_args,
            extra_link_args=openmp_flags),
    ],
    cmdclass={
        'sdist': SdistCommand,
        'build': BuildCommand,
        'bundle_client': BundleClientCommand,
        'build_ext': BuildExtCommand,
        'install': InstallCommand,
        'develop': DevelopCommand,
    },
)
